#include "stdafx.h"
#include "VirtualChannel.h"

#ifdef CAPTURE_FIRST_CHANNEL
 #include <stdio.h>
 static FILE *fp;
#endif

// Class vars...
static LPBYTE pSineWaveBuf = NULL;
static DWORD dwBufferSampleSize;
static DWORD dwSampleRate;
static WORD BytesPerSample;
static MODE mode;
static BITDEPTH DeviceBitDepth;
static bool bUseDeviceBitWidth;
static bool bInitializedClassVars = false;


VirtualChannel::VirtualChannel(int nChannel)
{
 bInitSuccess = false;
 bCreatedVirtualChn = false;
 pInputBuffer = NULL;

 if(!bInitializedClassVars)return; // You must call InitClassVars() before creating any channels!

 this->nChannel = nChannel;

 if(mode == MONO_PLAY_RECORD)
 {
    // We will save input data in a buffer to pass directly to output
    pInputBuffer = new BYTE [dwBufferSampleSize * BytesPerSample];
    if(pInputBuffer == NULL)return;
 }else if(mode == MONO_PLAY_ONLY)
 {
    // We will be sending a sinewave test tone to the output
	dwCurSampleOffset = 0;
 }else return; // mode must be one of the above!

#ifdef CAPTURE_FIRST_CHANNEL
 fp = fopen("out.pcm", "wb");
#endif

 bInitSuccess = true;
}

VirtualChannel::~VirtualChannel()
{
  if(bCreatedVirtualChn != NULL)
     CloseVirtualChannel(nChannel);
  if(pInputBuffer != NULL)
	 delete [] pInputBuffer;
  if(pSineWaveBuf != NULL)
  {
     delete [] pSineWaveBuf;
	 pSineWaveBuf = NULL;
  }

#ifdef CAPTURE_FIRST_CHANNEL
 if((nChannel == 0) && (fp != NULL))fclose(fp);
#endif
}

void VirtualChannel::PrepareVirtualChannel(void)
{
 if(bCreatedVirtualChn)return; // Only do this once
 ZeroMemory(&ChannelVars, sizeof(ChannelVars));
 ChannelVars.nVirtualChnNum = nChannel;
 ChannelVars.bUseDeviceBitWidthPlay = bUseDeviceBitWidth;
 ChannelVars.bUseDeviceBitWidthRecord = bUseDeviceBitWidth;
 ChannelVars.nBufSizeSamples = dwBufferSampleSize;
 ChannelVars.pOutBufServiceProc = RefillBuffer;
 ChannelVars.pInBufServiceProc = SaveBuffer;
 ChannelVars.dwInstanceData = (DWORD)(LPVOID)this;
 bCreatedVirtualChn = CreateVirtualChannel(&ChannelVars);
}

void VirtualChannel::PrepareBuffers(void)
{
 if(bCreatedVirtualChn)
 {
    if(mode == MONO_PLAY_RECORD)
    {
       ZeroMemory(pInputBuffer, dwBufferSampleSize * BytesPerSample);
    }else if(mode == MONO_PLAY_ONLY)
    {
	   dwCurSampleOffset = 0;
    }
    PreLoadBuffers(nChannel);
 }
}

void VirtualChannel::Start(void)
{
 if(bCreatedVirtualChn)
 {
    StartPlay(nChannel);
 }
}

void VirtualChannel::Stop(void)
{
 if(bCreatedVirtualChn)
 {
    StopPlay(nChannel);
 }
}

void SaveBuffer(LPBYTE pSndBuf, DWORD dwBufsizeBytes, DWORD dwInstanceData)
{
 LPVIRTUALCHANNEL pObj = (LPVIRTUALCHANNEL)(LPVOID)dwInstanceData;
 memcpy(pObj->pInputBuffer, pSndBuf, dwBufsizeBytes);
}

void RefillBuffer(LPBYTE pSndBuf, DWORD dwBufsizeBytes, DWORD dwInstanceData)
{
 LPVIRTUALCHANNEL pObj = (LPVIRTUALCHANNEL)(LPVOID)dwInstanceData;

 if(mode == MONO_PLAY_ONLY)
 {
    pObj->CopySineWaveOutput(pSndBuf, dwBufsizeBytes, &pObj->dwCurSampleOffset);
 }else // mode == MONO_PLAY_RECORD
 {
    // We are sending the input of this channel directly to its output... 
    memcpy(pSndBuf, pObj->pInputBuffer, dwBufsizeBytes);
 }

#ifdef CAPTURE_FIRST_CHANNEL
 // Capture whatever was just placed in pSndBuf
 if((pObj->nChannel == 0) && (fp != NULL))
    fwrite(pSndBuf, 1, dwBufsizeBytes, fp);
#endif
}
//*******************************************************************************



void VirtualChannel::InitClassVars(MODE modeReqd, bool bHiResSamples)
{
 if(bInitializedClassVars)return; // Already done!

 mode = modeReqd; 
 bUseDeviceBitWidth = bHiResSamples; 
 dwSampleRate = ASIOGetSampleRate();
 DeviceBitDepth = ASIOGetBitdepth();

 // Set default values for 16 bit samples...
 BytesPerSample = 2;

 // Now set bytes per sample according to our choices
 // (Not available in the free version)
 if(bUseDeviceBitWidth)
 {
    if(DeviceBitDepth == BITDEPTH_32BITS)
    {
	   BytesPerSample = 4;
	}else if(DeviceBitDepth == BITDEPTH_24BITS)
    {
	   BytesPerSample = 3;
	}
 }

 if(mode == MONO_PLAY_RECORD)
 {
    // We want minimal latency
    // since we are sending input directly to output
    SetMaxVirtualBuffers(2); // <--Has no effect on free version of ASIOHost
    // Go for the minimum buffer size permitted...
    dwBufferSampleSize = 2 * ASIOGetPrimaryBufferSizeSamples();
 }else
 {
    // We are only sending a sinewave test tone to the output,
    // so throughput latency is not a relevant issue.

    // The algorithm for assembling a sinewave signal from a single period
    // is dependent on a buffer size bigger than the period of the sinewave.

    // Calculate how many samples in 100 milliseconds at current sample rate...
    dwBufferSampleSize = ASIOGetSampleRate() / 10;
    // Now ensure we have an exact multiple of the current primary buffer size...
    DWORD dwPrimaryBufferSampleSize = ASIOGetPrimaryBufferSizeSamples();
    DWORD n = dwBufferSampleSize / dwPrimaryBufferSampleSize;
    if((dwBufferSampleSize % dwPrimaryBufferSampleSize) != 0)++n;
    if(n < 2)n = 2;
    dwBufferSampleSize = n * dwPrimaryBufferSampleSize;

    pSineWaveBuf = new BYTE [OUTDATA_BUFSIZE_SAMPLES * BytesPerSample];
    if(pSineWaveBuf == NULL)return;
    GetMonoSamples(pSineWaveBuf, SAMPLES_FOR_ONE_PERIOD);
 }

 bInitializedClassVars = true;
}

void VirtualChannel::GetMonoSamples(LPBYTE pSndBuf, DWORD dwSamples)
{
 double dSample;
 double dMaxAmplitude;
 double dPhaseAngle;
 double dRadians;
 double dFrequency;
 DWORD i, dwSampleRate = ASIOGetSampleRate();
 dFrequency = ((double)dwSampleRate / 100.0);
 // dFrequency = 441 Hz @ sample rate 44,100
 // dFrequency = 480 Hz @ sample rate 48,000
 // dFrequency = 960 Hz @ sample rate 96,000
 // etc.
 dRadians = TWO_PI * dFrequency / (double)dwSampleRate;
 dPhaseAngle = 0;

 if(bUseDeviceBitWidth)
 {
    if(DeviceBitDepth == BITDEPTH_32BITS)
    {
       long *pBuf = (long*)pSndBuf;
       dMaxAmplitude = Max32BitAmplitude;
	   for(i = 0; i < dwSamples; i++)
	   {
			dSample = dMaxAmplitude * sin(dPhaseAngle);
			if(dSample >= 0)
			{
			   *pBuf++ = (long)floor(dSample);
			}else
			{
			   *pBuf++ = (long)ceil(dSample);
			}
			dPhaseAngle += dRadians;
	   }
    }else if(DeviceBitDepth == BITDEPTH_24BITS)
    {
	   long sample;
       dMaxAmplitude = Max24BitAmplitude;
	   for(i = 0; i < dwSamples; i++)
	   {
			dSample = dMaxAmplitude * sin(dPhaseAngle);
			if(dSample >= 0)
			{
			   sample = (long)floor(dSample);
			}else
			{
			   sample = (long)ceil(dSample);
			}
            *pSndBuf++ = (sample) & 0xFF;
            *pSndBuf++ = (sample >> 8) & 0xFF;
            *pSndBuf++ = (sample >> 16) & 0xFF;
			dPhaseAngle += dRadians;
	   }
	}
 }else
 {
    short *pBuf = (short*)pSndBuf;
    dMaxAmplitude = Max16BitAmplitude;
    for(i = 0; i < dwSamples; i++)
    {
	    dSample = dMaxAmplitude * sin(dPhaseAngle);
	    if(dSample >= 0)
	    {
           *pBuf++ = (short)floor(dSample);
	    }else
	    {
           *pBuf++ = (short)ceil(dSample);
	    }
	    dPhaseAngle += dRadians;
    }
 }
}

// We are sending a sinewave test signal to the output,
// creating it on the fly by repeating the samples
// of a single period of the sinewave to the output buffer...

// Assumes at the start that dwSamplesReqd > SAMPLES_FOR_ONE_PERIOD
// dwCurSampleOffset is where we left off last time
void VirtualChannel::CopySineWaveOutput(LPBYTE pSndBuf, DWORD dwBufsizeBytes, LPDWORD pdwCurSampleOffset)
{
 LPBYTE pBuf = pSndBuf;
 DWORD dwSamplesReqd = dwBufferSampleSize; // same as dwBufsizeBytes / BytesPerSample;
 DWORD dwSamplesAvailable = SAMPLES_FOR_ONE_PERIOD - *pdwCurSampleOffset;
 DWORD dwBytes = BytesPerSample * dwSamplesAvailable;

 memcpy(pBuf, pSineWaveBuf + (*pdwCurSampleOffset * BytesPerSample), dwBytes);
 dwSamplesReqd -= dwSamplesAvailable;
 pBuf += dwBytes;
 *pdwCurSampleOffset = 0;
 dwBytes = BytesPerSample * SAMPLES_FOR_ONE_PERIOD;

 while(dwSamplesReqd >= SAMPLES_FOR_ONE_PERIOD)
 {
	   memcpy(pBuf, pSineWaveBuf, dwBytes);
	   dwSamplesReqd -= SAMPLES_FOR_ONE_PERIOD;
	   pBuf += dwBytes;
 }

 // Now dwSamplesReqd < dwSamplesAvailable
 if(dwSamplesReqd > 0)
 {
    memcpy(pBuf, pSineWaveBuf, BytesPerSample * dwSamplesReqd);
    *pdwCurSampleOffset = dwSamplesReqd;
 }
}
