// ASIOHostDemo.cpp : Defines the entry point for the application.
//
#include "stdafx.h"
#include <Windows.h>
#include "ASIOHost.h"
#include "VirtualChannel.h"
#include "resource.h"

#define MAX_LOADSTRING 100
#define MAXPATH 256
#define SCREEN_WIDTH 400
#define SCREEN_HEIGHT 200



// Free version is restricted to 16 bit samples
// Comment this out if you have the professional version of ASIOHost
// and you want to use full bitwidth offered by the device...
#define USE_16BIT_SAMPLES

// Free version is restricted to 2 channels
// Comment this out if you have the professional version of ASIOHost
// to test all channels of your sound card at once,
// or set to max channels desired...
#define MAX_CHANNELS 2


// Comment this out if you want inputs sent to outputs
// instead of just playing a sinewave test tone
#define OUTPUT_SINEWAVE_ONLY


// Global Variables:
HINSTANCE hInst;
HWND hwndApp;
TCHAR szTitle[MAX_LOADSTRING];
TCHAR szWindowClass[MAX_LOADSTRING];
char pPathname[MAXPATH];
char pCurrentDirPath[MAXPATH];

LPVIRTUALCHANNEL *ppVirtualChannels = NULL;
int nMaxChannels = 0;
int nChosenDevice = -1;
bool bASIOInitialized = false;

#ifdef USE_16BIT_SAMPLES
 bool bUseDeviceBitWidth = false;
#else
 bool bUseDeviceBitWidth = true;
#endif

#ifdef OUTPUT_SINEWAVE_ONLY
 MODE mode = MONO_PLAY_ONLY;
#else
 MODE mode = MONO_PLAY_RECORD;
#endif

ATOM				MyRegisterClass(HINSTANCE hInstance);
BOOL				InitInstance(HINSTANCE, int);
LRESULT CALLBACK	WndProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK	About(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK    SelectDevice(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);


void CreateVirtualChannels(void)
{
 nMaxChannels = ASIOGetMaxVirtualChannels(mode);

#ifdef MAX_CHANNELS
 if(nMaxChannels > MAX_CHANNELS)
 {
    nMaxChannels = MAX_CHANNELS;
 }
#endif

 if(nMaxChannels != ASIOCreateBuffers(nMaxChannels, mode))
 {
    return;
 }
 VirtualChannel::InitClassVars(mode, bUseDeviceBitWidth); // bUseDeviceBitWidth: Must pass false for free version of ASIOHost!
 ppVirtualChannels = new LPVIRTUALCHANNEL [nMaxChannels];
 if(ppVirtualChannels == NULL)return;
 ZeroMemory(ppVirtualChannels, sizeof(LPVIRTUALCHANNEL) * nMaxChannels);

 for(int i = 0; i < nMaxChannels; i++)
 {
     ppVirtualChannels[i] = new VirtualChannel(i);
	 if(ppVirtualChannels[i] != NULL)
	 {
	    if(!ppVirtualChannels[i]->GetInitSuccess())
	    {
           delete ppVirtualChannels[i];
	       ppVirtualChannels[i] = NULL;
	    }else
	    {
	       ppVirtualChannels[i]->PrepareVirtualChannel();
	    }
	 }
 }	 
}






int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
	MSG msg;
	HACCEL hAccelTable;

	LoadString(hInstance, IDS_APP_TITLE, szTitle, MAX_LOADSTRING);
	LoadString(hInstance, IDC_ASIOHOSTDEMO, szWindowClass, MAX_LOADSTRING);
	MyRegisterClass(hInstance);

	if (!InitInstance (hInstance, nCmdShow)) 
	{
		return FALSE;
	}

	hAccelTable = LoadAccelerators(hInstance, (LPCTSTR)IDC_ASIOHOSTDEMO);

	while (GetMessage(&msg, NULL, 0, 0)) 
	{
		if (!TranslateAccelerator(msg.hwnd, hAccelTable, &msg)) 
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	return msg.wParam;
}


ATOM MyRegisterClass(HINSTANCE hInstance)
{
	WNDCLASSEX wcex;

	wcex.cbSize = sizeof(WNDCLASSEX); 

	wcex.style			= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= (WNDPROC)WndProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= LoadIcon(hInstance, (LPCTSTR)IDI_ASIOHOSTDEMO);
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
	wcex.lpszMenuName	= MAKEINTRESOURCE(IDC_ASIOHOSTDEMO);
	wcex.lpszClassName	= szWindowClass;
	wcex.hIconSm		= LoadIcon(wcex.hInstance, (LPCTSTR)IDI_SMALL);

	return RegisterClassEx(&wcex);
}

BOOL InitInstance(HINSTANCE hInstance, int nCmdShow)
{
   HWND hWnd;

   hInst = hInstance;

   hWnd = CreateWindow(szWindowClass, szTitle, WS_OVERLAPPEDWINDOW,
      CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL);

   if (!hWnd)
   {
      return FALSE;
   }
   hwndApp = hWnd;
   ShowWindow(hWnd, nCmdShow);
   UpdateWindow(hWnd);

   return TRUE;
}

void CleanUp(void)
{
 if(ppVirtualChannels != NULL)
 {
    for(int i = 0; i < nMaxChannels; i++)
    {
	    if(ppVirtualChannels[i] != NULL)
	    {
		   delete ppVirtualChannels[i];
	    }
    }
	delete [] ppVirtualChannels;
	ppVirtualChannels = NULL;
 }
 ASIOShutdown();
}


LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	int wmId, wmEvent;
	PAINTSTRUCT ps;
	HDC hdc;

	switch (message) 
	{
		case WM_CREATE:
         ZeroMemory(pCurrentDirPath, MAXPATH);
		 GetCurrentDirectoryA(MAXPATH, pCurrentDirPath);
		 {
          int xborder = 2 * GetSystemMetrics(SM_CXFRAME);
          int yborder = 2 * GetSystemMetrics(SM_CYFRAME);
          int caption = GetSystemMetrics(SM_CYCAPTION);
          int menu = GetSystemMetrics(SM_CYMENU);
          SetWindowPos(hWnd, NULL, 0, 0, SCREEN_WIDTH + xborder, SCREEN_HEIGHT + yborder + caption + menu, SWP_NOCOPYBITS | SWP_NOMOVE | SWP_NOREPOSITION | SWP_NOZORDER);
		 }
 		break;
		case WM_WANTS_RESET:
         // Some sound cards require a complete reset when an application requests a change in sample rate
         // In this case, ASIOHost will send a WM_WANTS_RESET message, as defined in constants.h
         MessageBox(NULL, TEXT("Please exit the application and restart it.\0"),
                    TEXT("ASIO device reset request...\0"), MB_OK | MB_ICONEXCLAMATION);
		break;
		case WM_COMMAND:
			wmId    = LOWORD(wParam); 
			wmEvent = HIWORD(wParam); 
			// Parse the menu selections:
			switch (wmId)
			{
				case IDM_ABOUT:
				   DialogBox(hInst, (LPCTSTR)IDD_ABOUTBOX, hWnd, (DLGPROC)About);
				   break;
				case ID_VIEWDEVICES_SELECTDEVICE:
				   if((nChosenDevice >= 0) || bASIOInitialized)
				   {
		              MessageBox(NULL, TEXT("You already selected an ASIO device!"), TEXT("ASIO Devices..."), MB_OK);
				      break;
				   }
				   DialogBox(hInst, (LPCTSTR)IDD_DEVICES, hWnd, (DLGPROC)SelectDevice);
				   break;

				case IDM_FILE_START:
                 if((nChosenDevice < 0) || !bASIOInitialized)
                 {
                     MessageBox(NULL, TEXT("You need to select the device first!"), TEXT("ASIOHostDemo..."), MB_OK);
	                 break;
                 }
				 if(ppVirtualChannels != NULL)
				 {
                    int i;
                    for(i = 0; i < nMaxChannels; i++)
                    {
             	        if(ppVirtualChannels[i] != NULL)
	                    {
		                   ppVirtualChannels[i]->PrepareBuffers();
	                    }
                    }	 
                    for(i = 0; i < nMaxChannels; i++)
                    {
             	        if(ppVirtualChannels[i] != NULL)
	                    {
		                   ppVirtualChannels[i]->Start();
	                    }
                    }	 
				 }
				 break;

				case IDM_FILE_STOP:
                 if((nChosenDevice < 0) || !bASIOInitialized)
                 {
	                 break;
                 }
				 if(ppVirtualChannels != NULL)
				 {
                    for(int i = 0; i < nMaxChannels; i++)
                    {
             	        if(ppVirtualChannels[i] != NULL)
	                    {
		                   ppVirtualChannels[i]->Stop();
	                    }
                    }	 
				 }
				 break;

				case IDM_EXIT:
				 DestroyWindow(hWnd);
				   break;
				default:
				   return DefWindowProc(hWnd, message, wParam, lParam);
			}
			break;
		case WM_PAINT:
			hdc = BeginPaint(hWnd, &ps);
			EndPaint(hWnd, &ps);
			break;
		case WM_DESTROY:
			CleanUp();
			PostQuitMessage(0);
			break;
		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
   }
   return 0;
}

// Mesage handler for about box.
LRESULT CALLBACK About(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
		case WM_INITDIALOG:
				return TRUE;

		case WM_COMMAND:
			if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) 
			{
				EndDialog(hDlg, LOWORD(wParam));
				return TRUE;
			}
			break;
	}
    return FALSE;
}

// *Note - devices combo must not sort!
LRESULT CALLBACK SelectDevice(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
 HWND hWndCombo;
 char *pStrBuf, *ptr;
 int n, nDevices;

 switch (message)
 {
  case WM_INITDIALOG:
   pStrBuf = new char [2048];
   if(pStrBuf == NULL)return TRUE;
   ZeroMemory(pStrBuf, 2048);
   nDevices = ASIOGetDevices(pStrBuf, 2048);
   if(!nDevices)
   {
      delete [] pStrBuf;
	  return TRUE;
   }
   hWndCombo = GetDlgItem(hDlg, IDC_COMBO1);
   ptr = pStrBuf;
   for(n = 0; n < nDevices; n++)
   {
	   SendMessage(hWndCombo, CB_ADDSTRING, 0, (LPARAM) (LPCTSTR)ptr);
       // Device strings are separated by a NULL. Skip to next...
	   ptr += strlen(ptr);
	   ++ptr; 
   }
  delete [] pStrBuf;
  return TRUE;

  case WM_COMMAND:
   if(LOWORD(wParam) == IDOK) 
   {
      hWndCombo = GetDlgItem(hDlg, IDC_COMBO1);
	  nChosenDevice = SendMessage(hWndCombo, CB_GETCURSEL, 0, 0);
	  if(nChosenDevice >= 0)
	  {
		 // Now we have the desired device, go ahead and initialize ASIO...
		 bASIOInitialized = ASIOInit(hwndApp, nChosenDevice);
		 if(bASIOInitialized)
		 	CreateVirtualChannels();
	  }
	  EndDialog(hDlg, LOWORD(wParam));
	  return TRUE;
   }
   if(LOWORD(wParam) == IDCANCEL) 
   {
	  EndDialog(hDlg, LOWORD(wParam));
	  return TRUE;
   }
  break;
 }
 return FALSE;
}
